# -*- coding: utf-8 -*-
#import indago.optimizer as optimizer
#import indago.pso as pso
#import indago.fwa as fwa
#import indago.abca as abca

#import optimizer
#import pso
#import fwa
#import abca
__version__ = '0.1.2'
